"""
Functions to correct and filter data matrix from LC-MS Metabolomics data.
"""

import numpy as np
import pandas as pd
from scipy.interpolate import CubicSpline, interp1d
from statsmodels.nonparametric.smoothers_lowess import lowess
from typing import List, Callable, Union, Optional
from .utils import mad
from ._names import *


# def input_na(df: pd.DataFrame, classes: pd.Series, mode: str) -> pd.DataFrame:
#     """
#     Fill missing values.
#
#     Parameters
#     ----------
#     df : pd.DataFrame
#     classes: ps.Series
#     mode : {'zero', 'mean', 'min'}
#
#     Returns
#     -------
#     filled : pd.DataFrame
#     """
#     if mode == "zero":
#         return df.fillna(0)
#     elif mode == "mean":
#         return (df.groupby(classes)
#                 .apply(lambda x: x.fillna(x.mean()))
#                 .droplevel(0))
#     elif mode == "min":
#         return (df.groupby(classes)
#                 .apply(lambda x: x.fillna(x.min()))
#                 .droplevel(0))
#     else:
#         msg = "mode should be `zero`, `mean` or `min`"
#         raise ValueError(msg)


def average_replicates(data: pd.DataFrame, sample_id: pd.Series,
                       classes: pd.Series,
                       process_classes: List[str]) -> pd.DataFrame:
    """
    Group samples by id and computes the average.

    Parameters
    ----------
    data: pd.DataFrame
    sample_id: pd.Series
    classes: pd.Series
    process_classes: list[str]
    Returns
    -------
    pd.DataFrame
    """
    include_samples = classes[classes.isin(process_classes)].index
    exclude_samples = classes[~classes.isin(process_classes)].index
    mapper = sample_id[include_samples].drop_duplicates()
    mapper = pd.Series(data=mapper.index, index=mapper.values)
    included_data = data.loc[include_samples, :]
    excluded_data = data.loc[exclude_samples, :]
    averaged_data = (included_data.groupby(sample_id[include_samples])
                     .mean())
    averaged_data.index = averaged_data.index.map(mapper)
    result = pd.concat((averaged_data, excluded_data)).sort_index()
    return result


def correct_blanks(df: pd.DataFrame, classes: pd.Series,
                   corrector_classes: List[str], process_classes: List[str],
                   factor: float = 1.0, robust: bool = True,
                   mode: Union[str, Callable] = "mean",
                   process_blanks: bool = False) -> pd.DataFrame:
    """
    Correct samples using blanks.

    Parameters
    ----------
    df : pandas.DataFrame
        Data to correct.
    classes : pandas.Series
        Samples class labels.
    corrector_classes : list[str]
        Classes to be used as blanks.
    process_classes : list[str]
        Classes to be used as samples
    process_blanks : bool
        If True apply blank correction to corrector classes.
    factor : float
        factor used to convert low values to zero (see notes)
    mode : {'mean', 'max', 'lod', 'loq'} or function

    Returns
    -------
    corrected : pandas.DataFrame
        Data with applied correction

    """
    if robust:
        mean_func = lambda x: x.median()
        std_func = mad
    else:
        mean_func = lambda x: x.mean()
        std_func = lambda x: x.std()

    corrector = {"max": lambda x: x.max(),
                 "mean": lambda x: mean_func(x),
                 "lod": lambda x: mean_func(x) + 3 * std_func(x),
                 "loq": lambda x: mean_func(x) + 10 * std_func(x)}
    if hasattr(mode, "__call__"):
        corrector = mode
    else:
        corrector = corrector[mode]
    samples = df[classes.isin(process_classes)]
    blanks = df[classes.isin(corrector_classes)]

    correction = corrector(blanks)
    corrected = samples - correction
    corrected[(samples - factor * correction) < 0] = 0
    df[classes.isin(process_classes)] = corrected
    if process_blanks:
        corrected_blanks = blanks - correction
        corrected_blanks[(blanks - factor * correction) < 0] = 0
        df[classes.isin(corrector_classes)] = corrected_blanks
    return df


def _loocv_loess(x: pd.Series, y: pd.Series, interpolator: Callable,
                 frac: Optional[float] = None) -> tuple:
    """
    Helper function for batch_correction. Computes loess correction with LOOCV.

    Parameters
    ----------
    x: pd.Series
    y: pd.Series
    frac: float, optional
        fraction of sample to use in LOESS correction. If None, determines the
        best value using LOOCV.
    interpolator = callable
        interpolator function used to predict new values.
    Returns
    -------
    corrected: pd.Series
        LOESS corrected data
    """
    if frac is None:
        # valid frac values, from 4/N to 1/N, where N is the number of corrector
        # samples.
        frac_list = [k / x.size for k in range(4, x.size + 1)]
        rms = np.inf    # initial value for root mean square error
        best_frac = 1
        for frac in frac_list:
            curr_rms = 0
            for loocv_index in x.index[1:-1]:
                y_temp = y.drop(loocv_index)
                x_temp = x.drop(loocv_index)
                y_loess = lowess(y_temp, x_temp, return_sorted=False, frac=frac)
                interp = interpolator(x_temp, y_loess)
                curr_rms += (y[loocv_index] - interp(x[loocv_index])) ** 2
            if rms > curr_rms:
                best_frac = frac
                rms = curr_rms
        frac = best_frac
    return lowess(y, x, return_sorted=False, frac=frac)


def _generate_batches(df: pd.DataFrame, run_order: pd.Series, batch: pd.Series,
                      classes: pd.Series, corrector_classes: List[str],
                      process_classes: List[str]):
    batch_order = (pd.concat((batch, run_order), axis=1)
                   .sort_values([_sample_batch, _sample_order]))
    grouped = batch_order.groupby(_sample_batch)
    for n_batch, group in grouped:
        df_batch = df.loc[group.index, :]
        classes_batch = classes[group.index]
        process_df = df_batch.loc[classes_batch.isin(process_classes), :]
        corrector_df = df_batch.loc[classes_batch.isin(corrector_classes), :]
        process_order = run_order[process_df.index]
        corrector_order = run_order[corrector_df.index]
        batch_order = (run_order[corrector_df.index.union(process_df.index)]
                       .sort_values())
        corrector_df = corrector_df.set_index(corrector_order).sort_index()
        process_df = process_df.set_index(process_order).sort_index()
        yield corrector_df, process_df, batch_order


def get_outside_bounds_index(data: Union[pd.Series, pd.DataFrame], lb: float,
                             ub: float) -> pd.Index:
    """
    return index of columns with values outside bounds.
    Parameters
    ----------
    data: pd.Series or pd.DataFrame
    lb: float
        lower bound
    ub: float
        upper bound
    Returns
    -------

    """
    result = ((data < lb) | (data > ub))
    if isinstance(data, pd.DataFrame):
        result = result.all()
    if result.empty:
        return pd.Index([])
    else:
        return result[result].index


def batch_ext(order: pd.Series, batch: pd.Series, classes: pd.Series,
              class_list: List[str], ext: str) -> pd.Series:
    """
    get minimum/maximum order of samples of classes in class_list. Auxiliary
    function to be used with BatchChecker / FeatureCheckerBatchCorrection
    Parameters
    ----------
    order: pandas.Series
        run order
    batch: pandas.Series
        batch number
    classes: pandas.Series
        sample classes
    class_list: list[str]
        classes to be considered
    ext: {"min", "max"}
        Search for the min/max order in each batch.

    Returns
    -------
    pd.Series with the corresponding min/max order with batch as index.
    """
    func = {"min": lambda x: x.min(), "max": lambda x: x.max()}
    func = func[ext]

    ext_order = (order
                 .groupby([classes, batch])
                 .apply(func)
                 .reset_index()
                 .groupby(classes.name)
                 .filter(lambda x: x.name in class_list)
                 .groupby(batch.name)
                 .apply(func)[order.name])
    return ext_order


def check_qc_prevalence(data_matrix: pd.DataFrame,
                        batch: pd.Series, classes: pd.Series,
                        qc_classes: List[str], sample_classes: List[str],
                        threshold: float = 0,
                        min_qc_dr: float = 0.9) -> pd.Index:
    """
    Remove features with low detection rate in the QC samples. Also check that
    each feature is detected in the first and last block (this step is necessary
    interpolate the bias contribution to biological samples).

    Aux function to use in the BatchCorrector Pipeline.

    Parameters
    ----------
    data_matrix: DataFrame
    batch: Series
    classes: Series
    qc_classes: List[str]
    sample_classes: List[str]
    threshold: float
    min_qc_dr: float

    Returns
    -------
    index of invalid features
    """
    invalid_features = pd.Index([])
    for batch_number, batch_class in classes.groupby(batch):
        block_type, block_number = \
            make_sample_blocks(batch_class, qc_classes, sample_classes)
        qc_blocks = block_number[block_type == 0]
        block_prevalence = (data_matrix.loc[qc_blocks.index]
                            .groupby(qc_blocks)
                            .apply(lambda x: (x > threshold).any()))

        # check start block
        start_block_mask = block_prevalence.loc[qc_blocks.iloc[0]]
        tmp_rm = data_matrix.columns[~start_block_mask]
        invalid_features = invalid_features.union(tmp_rm)

        # check end block
        end_block_mask = block_prevalence.loc[qc_blocks.iloc[-1]]
        tmp_rm = data_matrix.columns[~end_block_mask]
        invalid_features = invalid_features.union(tmp_rm)

        # check qc prevalence
        n_blocks = qc_blocks.unique().size
        qc_prevalence = block_prevalence.sum() / n_blocks
        batch_min_qc_dr = max(4 / n_blocks, min_qc_dr)
        tmp_rm = data_matrix.columns[qc_prevalence < batch_min_qc_dr]
        invalid_features = invalid_features.union(tmp_rm)

    return invalid_features


def loess_interp(ft_data: pd.Series, order: pd.Series, qc_index: pd.Index,
                 sample_index: pd.Index, frac: float, interpolator: Callable,
                 n_qc: Optional[int] = None, method: str = "multiplicative"
                 ) -> pd.Series:
    """
    Applies LOESS-correction interpolation on a feature. Auxiliary function of
    batch_corrector_func

    Parameters
    ----------
    ft_data: pd.Series
        Feature intensity
    order: pd.Series
    qc_index: pd.Index
    sample_index: pd.Index
    frac: float
    interpolator: Callable
    n_qc: int, optional
    Number of QCs involved in mean calculation. If None, all QCs are involved.
    method : {"multiplicative", "additive"}
        Method used to model variation.
    Returns
    -------
    pd.Series
    """
    if n_qc is None:
        n_qc = qc_index.size

    qc_median = ft_data[qc_index[:n_qc]].median()
    # if there are several 0 values the median may be 0, this prevent against
    # such cases
    if np.isclose(qc_median, 0.0):
        qc_median = ft_data[qc_index[:n_qc]].mean()

    qc_loess = _loocv_loess(order[qc_index], ft_data[qc_index], interpolator,
                            frac=frac)
    interp = interpolator(order[qc_index], qc_loess)

    if method == "additive":
        bias = interp(order[sample_index]) - qc_median
        ft_data[sample_index] -= bias
    elif method == "multiplicative":
        qc_smooth = interp(order[sample_index])
        # smoothed values close to zero can become negative. As we are using
        # a multiplicative correction, this prevents changing signs while
        # applying the correction
        if qc_median <= 0.0:
            factor = 0 * qc_smooth
        else:
            qc_smooth[qc_smooth <= 0] = qc_median
            factor = qc_median / qc_smooth
        ft_data[sample_index] *= factor
    else:
        msg = "Valid methods are `additive` or `multiplicative`."
        raise ValueError(msg)
    return ft_data


def batch_corrector_func(df_batch: pd.DataFrame, order: pd.Series,
                         classes: pd.Series, frac: float,
                         interpolator: Callable, qc_classes: List[str],
                         sample_classes: List[str],
                         n_qc: Optional[int] = None,
                         method: str = "multiplicative") -> pd.DataFrame:
    """
    Applies LOESS correction - interpolation on a single batch. Auxiliary
    function of interbatch_correction.

    Parameters
    ----------
    df_batch: pandas.DataFrame
    order: pandas.Series
    classes: pandas.Series
    frac: float
    interpolator: Callable
    qc_classes: list[str]
    sample_classes: list[str]
    n_qc: int, optional
    Number of QCs involved in mean calculation. If None, all QCs are involved.
    method : {"multiplicative", "additive"}
        Method used to model variation.

    Returns
    -------
    pandas.DataFrame
    """
    qc_index = classes.isin(qc_classes)
    qc_index = qc_index[qc_index].index
    sample_index = classes.isin(sample_classes)
    sample_index = sample_index[sample_index].index
    df_batch.loc[sample_index, :] = \
        (df_batch.apply(loess_interp,
                        args=(order, qc_index, sample_index, frac,
                              interpolator),
                        n_qc=n_qc, method=method))
    return df_batch


def interbatch_correction(df: pd.DataFrame, order: pd.Series, batch: pd.Series,
                          classes: pd.Series, corrector_classes: List[str],
                          process_classes: List[str],
                          frac: Optional[float] = None,
                          interpolator: Optional[str] = "splines",
                          n_qc: Optional[int] = None,
                          process_qc: bool = True,
                          method: str = "multiplicative"
                          ) -> pd.DataFrame:
    r"""
    Correct instrument response drift using LOESS regression [1, 2]
    and center each batch to a common mean.

    Parameters
    ----------
    df : pandas.DataFrame
    order : pandas.Series
        run order of samples
    batch: pandas.Series
        batch number of samples
    classes : pandas.Series
        class label for samples
    corrector_classes : str
       label of corrector class
    process_classes: list[str]
       samples to correct
    frac: float, optional.
       fraction of samples used to build local regression.If None, finds the
       best value using LOOCV.
    interpolator: {"linear", "splines"}
        Type of interpolator to use.
    n_qc: int, optional
        Number of QCs involved in mean calculation. If None, all QCs are
        involved.
    process_qc : bool
        If True, applies correction to QC samples.
    method : {"multiplicative", "additive"}
        Method used to model variation.

    Returns
    -------
    corrected: pandas.DataFrame
       corrected data
    invalid_features: pandas.Series
       features that where corrected in a number of batches lower than
       min_batch_prevalence.

    References.
    -----
    .. [1] W B Dunn *et al*, "Procedures for large-scale metabolic profiling of
    serum and plasma using gas chromatography and liquid chromatography coupled
    to mass spectrometry", Nature Protocols volume 6, pages 1060–1083 (2011).
    .. [2] D Broadhurst *et al*, "Guidelines and considerations for the use of
    system suitability and quality control samples in mass spectrometry assays
    applied in untargeted clinical metabolomic studies.", Metabolomics,
    2018;14(6):72. doi: 10.1007/s11306-018-1367-3

    Notes
    -----
    The correction is applied as described by Broadhurst in [2]. Using QC
    samples a correction is generated for each feature in the following way:
    The signal of a Quality control can be described in terms of three
    components: a mean value, a systematic bias f and error.

    .. math::
       m_{i} = \bar{m_{i}} + f(t) + \epsilon

    f(t) is estimated after mean subtraction using Locally weighted scatter plot
    smoothing (LOESS). The optimal fraction of samples for each local
    regression is found using LOOCV.

    Mean centering is performed subtracting a batch mean and adding a grand
    mean.
    """
    interp_func = {"splines": CubicSpline, "linear": interp1d}
    interp_func = interp_func[interpolator]

    if process_qc:
        # add QC classes to process classes
        process_classes = corrector_classes + process_classes
        process_classes = list(set(process_classes))

    def corrector_helper(df_group):
        return batch_corrector_func(df_group, order[df_group.index],
                                    classes[df_group.index], frac,
                                    interp_func, corrector_classes,
                                    process_classes, n_qc=n_qc,
                                    method=method)

    # intra batch correction
    corrected = df.groupby(batch).apply(corrector_helper)

    # inter batch mean alignment
    def batch_mean_func(df_group):
        batch_mean = (df_group[classes[df_group.index].isin(corrector_classes)]
                      .mean())
        is_process_sample = classes[df_group.index].isin(process_classes)
        if method == "additive":
            df_group[is_process_sample] -= batch_mean
        else:
            batch_mean[batch_mean <= 0] = np.nan
            df_group[is_process_sample] /= batch_mean
            df_group.fillna(0)
        return df_group

    global_median = corrected[classes.isin(corrector_classes)].median()
    corrected = corrected.groupby(batch).apply(batch_mean_func)
    process_mask = classes.isin(process_classes)
    if method == "additive":
        corrected.loc[process_mask, :] += global_median
    else:
        corrected.loc[process_mask, :] *= global_median
    corrected[corrected < 0] = 0

    return corrected


def make_sample_blocks(classes: pd.Series, corrector_classes: List[str],
                       process_classes: List[str]):
    """
    groups samples into blocks of consecutive samples of the same type
    aux function in BatchCorrector pipeline.

    each class is assigned to each one of three possible sample blocks:
    0 if the sample is mapped as QC, 1 if the sample is mapped as a
    sample, and 2 otherwise.

    Each block is assigned an unique number.
    """
    class_to_block_type = dict()
    for c in classes.unique():
        if c in corrector_classes:
            class_to_block_type[c] = 0
        elif c in process_classes:
            class_to_block_type[c] = 1
        else:
            class_to_block_type[c] = 2
    block_type = classes.map(class_to_block_type)
    block_number = (block_type.diff().fillna(0) != 0).cumsum()
    return block_type, block_number



import numpy as np
import pandas as pd
from typing import List, Optional
from .container import DataContainer


def simulate_dataset(population: dict, mean_dict: dict, cov_dict: dict,
                     feature_mz: np.ndarray, feature_rt: np.ndarray,
                     blank_contribution: Optional[np.ndarray] = None,
                     noise_dict: Optional[dict] = None,
                     qc_template: bool = True, qc_bracket_size: int = 5,
                     triple_qc: bool = True, prepend_blank: int = 0,
                     append_blank: int = 0, batch_size: int = 25
                     ) -> DataContainer:
    """
    Creates DataContainer with simulated data

    Parameters
    ----------
    population: dict[str, int]
        A dictionary where the keys are the available classes in
        the experiment and the values are the number of samples
        from each class.
    mean_dict : dict
        a dictionary where each key is a class and the value is a 1D array with
        the mean values of each feature in the class
    cov_dict : dict
        a dictionary where each key is a class and the value is an array with
        the covariance matrix for the features in each class. A 2D array must
        be used to create correlated data. a 1D array where each element is
        the variance of the features can be used to create uncorrelated data.
    feature_mz : array
        m/z values assigned to each feature.
    feature_rt : array
        Rt values assigned to each feature.
    noise_dict : dict
        a dictionary where each key is a class and the value is a 1D array
        with the standard deviation of an additive noise level for each feature.
    blank_contribution : array, optional
        an with the mean value for each feature in the blanks.
    qc_template: bool
        If True, includes a template of QC samples surrounding
        the study samples.
    qc_bracket_size: int
        The number of study samples between QCs samples.
    triple_qc: bool
        If True adds three consecutive QC samples before the study samples
    prepend_blank: int
        number of blank samples to prepend to the batch
    append_blank: int`
        number of blank samples to append to the batch
    batch_size : int
        Number of study samples per analytical batch.

    Returns
    -------
    DataContainer

    """

    sm = _make_sample_list(population, qc_template, qc_bracket_size,
                           triple_qc, prepend_blank, append_blank, batch_size)
    dm = _make_data_matrix(population, sm, mean_dict, cov_dict,
                           blank_contribution, noise_dict)
    fm = np.vstack((feature_mz, feature_rt)).T
    fm = pd.DataFrame(data=fm, columns=["mz", "rt"], index=dm.columns)
    mapping = {"sample": list(population)}
    if qc_template:
        mapping["qc"] = ["QC"]
    if max(prepend_blank, append_blank) > 0:
        mapping["blank"] = ["blank"]
    return DataContainer(dm, fm, sm, mapping)


def _make_sample_list(population: dict, qc_template: bool = True,
                      qc_bracket_size: int = 5, triple_qc: bool = True,
                      prepend_blank: int = 0, append_blank: int = 0,
                      batch_size: int = 25):
    """
    creates a simulated sample list for an untargeted metabolomic assay.

    Parameters
    ----------
    population: dict[str, int]
        A dictionary where the keys are the available classes in
        the experiment and the values are the number of samples
        from each class.
    qc_template: bool
        If True, includes a template of QC samples surrounding
        the study samples.
    qc_bracket_size: int
        The number of study samples between QCs samples.
    triple_qc: bool
        If True adds three consecutive QC samples before the study samples
    prepend_blank: int
        number of blank samples to prepend to the batch
    append_blank: int`
        number of blank samples to append to the batch
    batch_size : int
        Number of study samples per analytical batch.

    """
    n_study_samples = sum(population.values())

    prepend_blank = ["blank"] * prepend_blank
    append_blank = ["blank"] * append_blank

    # append QCs before and after study samples
    if qc_template:
        if triple_qc:
            n_append = 3
        else:
            n_append = 1
        prepend_blank = prepend_blank + ["QC"] * n_append
        append_blank = ["QC"] * n_append + append_blank
    else:
        # setting this bracket size will avoid including QC samples
        qc_bracket_size = n_study_samples + 1

    # randomize classes and make QC brackets
    classes = _make_sample_classes(population, batch_size,
                                   qc_bracket_size, prepend_blank, append_blank)
    n_prepend = len(prepend_blank)
    n_append = len(append_blank)
    batch_number = _make_batch_number(batch_size, n_study_samples, n_prepend,
                                      n_append, qc_bracket_size)
    order = list(range(1, len(batch_number) + 1))
    sample_name = _make_sample_name(len(batch_number))
    sample_list = {"id": sample_name, "class": classes, "order": order,
                   "batch": batch_number}
    sample_list = pd.DataFrame(data=sample_list, index=sample_name)
    sample_list.index.name = "samples"
    return sample_list


def _make_data_matrix(population: dict, sample_list: pd.DataFrame,
                      mean_dict: dict, cov_dict: dict,
                      blank_contribution: Optional[np.ndarray] = None,
                      noise_dict: Optional[dict] = None):
    """
    Creates a simulated data matrix

    Parameters
    ----------
    population: dict
        A dictionary where each key is a sample class and the value is the
        number of samples.
    sample_list : DataFrame
        sample list created with _make_sample_list
    mean_dict : dict
        a dictionary where each key is a class and the value is a 1D array with
        the mean values of each feature in the class
    cov_dict : dict
        a dictionary where each key is a class and the value is an array with
        the covariance matrix for the features in each class. A 2D array must
        be used to create correlated data. a 1D array where each element is
        the variance of the features can be used to create uncorrelated data.
    noise_dict : dict
        a dictionary where each key is a class and the value is a 1D array
        with the standard deviation of an additive noise level for each feature.
    blank_contribution : array, optional
        an with the mean value for each feature in the blanks.

    Returns
    -------

    """
    data_mats = dict()
    for k_class, k_obs in population.items():
        mean = mean_dict[k_class]
        cov = cov_dict[k_class]
        mat = generate_correlated_data(k_obs, mean, cov)
        data_mats[k_class] = mat

    if (sample_list["class"] == "blank").any():
        if blank_contribution is None:
            blank_contribution = np.zeros_like(mean)
        data_mats["blank"] = \
            _generate_blank_data(sample_list, blank_contribution)

    if (sample_list["class"] == "QC").any():
        data_mats["QC"] = _generate_qc_data(population, sample_list,
                                            mean_dict)

    if noise_dict is not None:
        for c, x in data_mats.items():
            data_mats[c] = _add_noise(x, noise_dict[c])

    data = _concat_data(data_mats, sample_list)
    ft_names = _generate_feature_names(data.shape[1])
    data_matrix = pd.DataFrame(data=data, index=sample_list.index,
                               columns=ft_names)
    return data_matrix


def _make_sample_classes(sample_distribution, batch_size, qc_bracket_size,
                         prepend, append):
    n_study_samples = sum(sample_distribution.values())
    # randomize classes and make QC brackets
    classes = list()
    for k, v in sample_distribution.items():
        classes.extend([k] * v)

    permutation_index = np.random.permutation(n_study_samples)
    qb, rb = divmod(n_study_samples, batch_size)
    n_batches = qb + bool(rb)
    res = list()
    start = 0
    end = min(batch_size, n_study_samples)
    for k_batches in range(n_batches):
        res += prepend
        res += _make_batch_sample_block(classes,
                                        permutation_index[start:end],
                                        qc_bracket_size)
        res += append
        start = end
        end = min(start + batch_size, n_study_samples)
    return res


def _make_batch_number(batch_size, n_study_samples, n_prepend, n_append,
                       qc_bracket_size):
    n_batch, n_study_samples_last_batch = divmod(n_study_samples, batch_size)
    # adds an extra batch for the last batch
    n_batch += int(n_study_samples_last_batch > 0)
    n_qc_batch = _n_qc_per_batch(batch_size, qc_bracket_size)
    n_samples_per_batch = batch_size + n_prepend + n_append + n_qc_batch
    batch_number = list()
    for k_batch in range(1, n_batch):
        batch_number.extend([k_batch] * n_samples_per_batch)
    if n_study_samples_last_batch:

        n_qc_last_batch = _n_qc_per_batch(n_study_samples_last_batch,
                                          qc_bracket_size)
        n_samples_last_batch = (n_study_samples_last_batch + n_prepend +
                                n_append + n_qc_last_batch)
        batch_number.extend([n_batch] * n_samples_last_batch)
    return batch_number


def _make_sample_name(n_samples: int):
    pad_length = len(str(n_samples))
    template_str = "Sample-{{:0{}n}}".format(pad_length)
    sample_name = [template_str.format(x) for x in range(1, n_samples + 1)]
    return sample_name


def generate_correlated_data(n_obs: int, mean: np.array,
                             cov: np.array) -> np.array:
    """
    Generates a data matrix with normal correlated columns

    Parameters
    ----------
    n_obs : int
        number of observations in the data matrix
    mean : vector with length `n_vars`
        mean vector of the variables.
    cov : vector with length `n_vars` or matrix with shape [n_vars, n_vars]
        covariance matrix of the variables. If `cov` is a vector, then
        a diagonal covariance matrix is used.

    Returns
    -------
    X : array with shape [n_obs, n_vars]
        a matrix with rows of observations and columns of variables

    """
    if len(cov.shape) == 1:
        cov = np.diag(cov)

    if not _is_valid_covariance_matrix(cov):
        msg = "`cov` must be symmetric and positive definite."
        raise ValueError(msg)

    if cov.shape[1] != mean.size:
        msg = "Shape mismatch between the mean and the covariance matrix"
        raise ValueError(msg)

    L = np.linalg.cholesky(cov)
    n_vars = mean.size
    X = np.random.normal(size=(n_obs, n_vars))
    X = np.matmul(X, L.T) + mean
    return X


def _generate_blank_data(sample_list, blank_contribution):
    n_blank = (sample_list["class"] == "blank").sum()
    blank_data = np.tile(blank_contribution, [n_blank, 1])
    return blank_data


def _generate_qc_data(sample_population, sample_list, mean_dict):
    n_qc = (sample_list["class"] == "QC").sum()
    n_samples = sum(sample_population.values())
    mean = sum([n * mean_dict[c] for c, n in sample_population.items()])
    mean = np.array(mean) / n_samples
    qc_data = np.tile(mean, [n_qc, 1])
    return qc_data


def _is_valid_covariance_matrix(cov: np.array):
    """Checks if a matrix is symmetric and positive definite"""
    is_sym = _is_symmetric(cov)
    is_def_pos = (np.linalg.eigvals(cov) > 0).all()
    return is_sym and is_def_pos


def _is_symmetric(x: np.array) -> bool:
    """Check if a matrix is symmetric"""
    return (x == x.T).all()


def _generate_random_covariance_matrix(n: int) -> np.array:
    """Creates a random symmetric, positive definite matrix"""
    x = np.random.normal(size=(n, n))
    x = (x + x.T) / 2 + np.diag(np.abs(x).sum(axis=1)) * 2
    return x


def _make_batch_sample_block(classes, batch_perm_index,
                             qc_bracket_size):
    classes_rand = list()
    for k, ind in enumerate(batch_perm_index):

        if ((k % qc_bracket_size) == 0) and k > 0:
            classes_rand.append("QC")
        classes_rand.append(classes[ind])
    return classes_rand


def _n_qc_per_batch(batch_size, qc_bracket_size):
    q, r = divmod(batch_size, qc_bracket_size)
    if q > 0 and r == 0:
        q -= 1
    return q


def _add_noise(x, noise):
    xn = x + np.random.normal(size=x.shape[1], scale=noise)
    xn[xn < 0] = 0
    return xn


def _concat_data(data_mats, sample_list):
    classes = sample_list["class"]
    classes_unique = np.sort(classes.unique())
    mat = [data_mats[x] for x in classes_unique]
    mat = np.vstack(mat)
    # sort in sample list order
    sorted_index = np.argsort(classes).values
    data = np.zeros_like(mat)
    data[sorted_index, :] = mat
    return data


def _generate_feature_names(n_ft):
    pad_length = len(str(n_ft))
    template_str = "FT-{{:0{}n}}".format(pad_length)
    ft_name = [template_str.format(x) for x in range(1, n_ft + 1)]
    return ft_name



"""
functions and objects used to detect peaks.

Objects
-------

- Peak : Stores peak location and extension. Computes peak parameters.

Functions
---------
- estimate_noise(x) : Estimates noise level in a 1D signal
- estimate_baseline(x, noise) : Estimates the baseline in a 1D signal
- detect_peaks(x, y) : Detects peaks in a 1D signal
- find_centroids(x, y) : Computes the centroid and area of peaks in a 1D signal.

"""

import numpy as np
from scipy.integrate import trapz
from scipy.integrate import cumtrapz
from scipy.interpolate import interp1d
from scipy.signal import find_peaks
from scipy.special import erfc
from scipy.stats import median_abs_deviation as mad
from typing import Callable, Dict, List, Optional, Tuple


class Peak:
    """
    Representation of a peak. Computes peak descriptors.

    Attributes
    ----------
    start: int
        index where the peak begins. Must be smaller than `apex`
    apex: int
        index where the apex of the peak is located. Must be smaller than `end`
    end: int
        index where the peak ends. Start and end used as slices defines the
        peak region.

    """

    def __init__(self, start: int, apex: int, end: int):

        try:
            assert start < apex
            assert apex < end
        except AssertionError:
            msg = "start must be lower than loc and loc must be lower than end"
            raise InvalidPeakException(msg)

        self.apex = apex
        self.start = start
        self.end = end

    def __repr__(self):
        str_repr = "{}(start={}, apex={}, end={})"
        name = self.__class__.__name__
        str_repr = str_repr.format(name, self.start, self.apex, self.end)
        return str_repr

    def get_loc(self, x: np.ndarray, y: np.ndarray) -> float:
        """
        Finds the peak location in x, using y as weights.

        Parameters
        ----------
        x : sorted array
        y : array with the same size as x

        Returns
        -------
        loc : float

        """
        weights = y[self.start:self.end]
        weights[weights < 0] = 0
        loc = np.abs(np.average(x[self.start:self.end], weights=weights))
        return loc

    def get_height(self, y: np.ndarray, baseline: np.ndarray) -> float:
        """
        Computes the height of the peak, defined as the difference between the
        value of y and the baseline at the peak apex.

        Parameters
        ----------
        y : array
        baseline : array with the same size as y

        Returns
        -------
        height : non-negative number. If the baseline estimation is greater
        than y, the height is set to zero.

        """
        height = y[self.apex] - baseline[self.apex]
        return max(0.0, height)

    def get_area(self, x: np.ndarray, y: np.ndarray, baseline: np.ndarray
                 ) -> float:
        """
        Computes the area in the region defined by the peak.

        If the baseline area is greater than the peak area, the area is set
        to zero.

        Parameters
        ----------
        x : sorted array
        y : array with the same size as y
        baseline : array with the same size as y

        Returns
        -------
        area : positive number.

        """
        baseline_corrected = (y[self.start:self.end] -
                              baseline[self.start:self.end])
        area = trapz(baseline_corrected, x[self.start:self.end])
        return max(0.0, area)

    def get_width(self, x: np.ndarray, y: np.ndarray, baseline: np.ndarray
                  ) -> float:
        """
        Computes the peak width, defined as the region where the 95 % of the
        total peak area is distributed.

        Parameters
        ----------
        x : sorted array
        y : array with the same size as y
        baseline : array with the same size as y

        Returns
        -------
        width : positive number.

        """
        height = (y[self.start:self.end] -
                  baseline[self.start:self.end])
        area = cumtrapz(height, x[self.start:self.end])
        if area[-1] > 0:
            relative_area = area / area[-1]
            percentile = [0.025, 0.975]
            start, end = self.start + np.searchsorted(relative_area, percentile)
            width = x[end] - x[start]
        else:
            width = 0.0
        return max(0.0, width)

    def get_extension(self, x: np.ndarray) -> float:
        """
        Computes the peak extension, defined as the length of the peak region.

        Parameters
        ----------
        x: sorted array

        Returns
        -------
        extension : positive number

        """
        return x[self.end] - x[self.start]

    def get_snr(self, y: np.array, noise: np.array, baseline: np.array
                ) -> float:
        """
        Computes the peak signal-to-noise ratio, defined as the quotient
        between the peak height and the noise level at the apex.

        Parameters
        ----------
        y : array
        noise : array with the same size as y
        baseline : array with the same size as y

        Returns
        -------
        snr : float

        """

        peak_noise = noise[self.apex]
        if np.isclose(peak_noise, 0):
            snr = np.inf
        else:
            snr = self.get_height(y, baseline) / peak_noise
        return snr

    def get_descriptors(self, x: np.array, y: np.array, noise: np.array,
                        baseline: np.array) -> Dict[str, float]:
        """
        Computes peak height, area, location, width and SNR.

        Parameters
        ----------
        x : sorted array
        y : array with the same size as x
        noise : array with the same size as x
        baseline : array with the same size as x

        Returns
        -------
        descriptors: dict
            A mapping of descriptor names to descriptor values.
        """
        descriptors = {"height": self.get_height(y, baseline),
                       "area": self.get_area(x, y, baseline),
                       "loc": self.get_loc(x, y),
                       "width": self.get_width(x, y, baseline),
                       "snr": self.get_snr(y, noise, baseline)}
        return descriptors


def detect_peaks(
        x: np.ndarray,
        noise: np.ndarray,
        baseline: np.ndarray,
        find_peaks_params: Optional[dict] = None
) -> List[Peak]:
    r"""
    Finds peaks in a 1D signal.

    Parameters
    ----------
    x : array
        Signal with peaks.
    noise : array with the same size as x
        Noise level of x.
    baseline : array with the same size as x
        Baseline estimation of x
    find_peaks_params : dict or None, default=None
        parameters to pass to :py:func:`scipy.signal.find_peaks`.

    Returns
    -------
    peaks : List[Peak]
        list of detected peaks

    Notes
    -----
    The algorithm for peak finding is as follows:

    1.  Peaks are detected using :py:func:`scipy.signal.find_peaks`. Peaks
        with a prominence lower than three times the noise or in regions
        classified as baseline  are removed.
    2.  Points from :math:`x` are considered baseline is the following
        condition is meet:

        .. math::
            |x[k] - b[k]| < e[k]

        where :math:`b` is the baseline and :math:`e` is the noise. If a
        detected peak is classified as baseline is removed.
    3.  The extension of each peak is found by finding the closest baseline
        point to its left and right.
    4.  If there are overlapping peaks (i.e. overlapping peak extensions),
        the extension is fixed by defining a boundary between the peaks as
        the minimum value between the two peaks.

    See Also
    --------
    estimate_noise : estimates noise in a 1D signal
    estimate_baseline : estimates the baseline in a 1D signal
    Peak : stores peak start, apex and end indices
    get_peak_descriptors : computes descriptors on a list of Peak objects

    """

    baseline_index = np.where((x - baseline) < noise)[0]
    prominence = 3 * noise

    if find_peaks_params is None:
        find_peaks_params = {"prominence": prominence, "distance": 3}
    else:
        find_peaks_params["prominence"] = prominence

    peaks = find_peaks(x, **find_peaks_params)[0]
    # remove peaks close to baseline level
    peaks = np.setdiff1d(peaks, baseline_index, assume_unique=True)

    start, end = _find_peak_extension(peaks, baseline_index)
    start, end = _fix_peak_overlap(x, start, peaks, end)
    start, peaks, end = _normalize_peaks(x, start, peaks, end)
    peaks = [Peak(s, p, e) for s, p, e in zip(start, peaks, end)]
    return peaks


def get_peak_descriptors(
        x: np.ndarray,
        y: np.ndarray,
        noise: np.ndarray,
        baseline: np.ndarray,
        peaks: List[Peak],
        descriptors: Optional[Dict[str, Callable]] = None,
        filters: Optional[Dict[str, Tuple]] = None
) -> Tuple[List[Peak], List[Dict[str, float]]]:
    """
    Computes peak descriptors for a list of peaks.

    By default, the location, height, area, width and SNR of each peak is
    computed, but custom descriptors can also be included.

    Parameters
    ----------
    x : sorted array
    y : array with the same size as x
        Signal where the peaks were detected.
    noise: array with the same size as x.
        Noise estimation of y.
    baseline: array with the same size as x.
        Baseline estimation of y.
    peaks : List[Peaks]
    descriptors : dict or None, default=None
        A dictionary of strings to callables, used to estimate custom
        descriptors of a peak. The function must have the following signature:

        .. code-block:: python

            "estimator_func(x, y, noise, baseline, peak) -> float"

    filters : dict or None, default=None
        A dictionary of descriptor names to a tuple of minimum and maximum
        acceptable values. To use only minimum/maximum values, use None
        (e.g. (None, max_value) in the case of using only maximum).Peaks with
        descriptors outside these ranges are removed. Filters for custom
        descriptors can also be used.

    Returns
    -------
    peaks : List[Peak]
        filtered list of peaks
    descriptors: List[dict]
        Descriptors for each peak.

    See Also
    --------
    detect_peaks : detects peaks in a 1D signal

    """

    if descriptors is None:
        descriptors = dict()

    if filters is None:
        filters = dict()
    _fill_filter_boundaries(filters)

    valid_peaks = list()
    descriptor_list = list()
    for p in peaks:
        p_descriptors = p.get_descriptors(x, y, noise, baseline)
        for descriptor, func in descriptors.items():
            p_descriptors[descriptor] = func(x, y, noise, baseline, p)

        if _has_all_valid_descriptors(p_descriptors, filters):
            valid_peaks.append(p)
            descriptor_list.append(p_descriptors)
    return valid_peaks, descriptor_list


def estimate_noise(
        x: np.ndarray,
        min_slice_size: int = 200,
        n_slices: int = 5,
        robust: bool = True
) -> np.ndarray:
    """
    Estimates the noise level in a signal.

    Splits x into several slices and estimates the noise assuming that the
    noise is gaussian iid in each slice. See [ADD LINK] for a detailed
    description of how the method works

    Parameters
    ----------
    x : 1D array
    min_slice_size : int, default=200
        Minimum size of a slice. If the size of x is smaller than this value,
        the noise is estimated using the whole array.
    n_slices: int, default=5
        Number of slices to create. The size of each slice must be greater than
        `min_slice_size`.
    robust : bool, default=True
        If True, estimates the noise using the median absolute deviation. Else
        uses the standard deviation.

    Returns
    -------
    noise: array with the same size as x

    """
    noise = np.zeros_like(x)
    slice_size = x.size // n_slices
    if slice_size < min_slice_size:
        slice_size = min_slice_size
    start = 0
    while start < x.size:
        end = min(start + slice_size, x.size)
        if (x.size - end) < min_slice_size:
            # prevent short slices at the end of x
            end = x.size
        slice_noise = _estimate_local_noise(x[start:end], robust=robust)
        noise[start:end] = slice_noise
        start = end
    return noise


def estimate_baseline(
        x: np.ndarray,
        noise: np.ndarray,
        min_proba: float = 0.05
) -> np.ndarray:
    """
    Computes the baseline of a 1D signal.

    The baseline is estimated by classifying each point in the signal as either
    signal or baseline. The baseline is obtained by interpolation of baseline
    points. See [ADD LINK] for a detailed explanation of how the method works.

    Parameters
    ----------
    x : non-empty 1D array
    noise : array
        Noise estimation obtained with ``estimate_noise``
    min_proba : number between 0 and 1, default=0.05

    Returns
    -------
    baseline : array with the same size as x

    """
    # find points that only have contribution from the baseline
    baseline_index = _find_baseline_points(x, noise, min_proba)

    # interpolate baseline points to match x size
    baseline = x[baseline_index]
    interpolator = interp1d(baseline_index, baseline)
    baseline = interpolator(np.arange(x.size))

    # prevents that interpolated points have higher values than x.
    baseline = np.minimum(baseline, x)
    return baseline


def find_centroids(mz: np.ndarray, spint: np.ndarray, min_snr: float,
                   min_distance: float
                   ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Finds the centroid of a mass spectrum in profile mode.

    Parameters
    ----------
    mz : array
    spint : array
    min_snr : positive number
        Minimum signal-to-noise ratio
    min_distance : positive number
        Minimum m/z distance between consecutive centroids

    Returns
    -------
    centroid_mz : array
        centroid m/z of peaks
    centroid_int : array
        area of peaks

    """
    noise = estimate_noise(spint)
    baseline = estimate_baseline(spint, noise)
    baseline_index = np.where((spint - baseline) < noise)[0]
    prominence = 3 * noise
    find_peaks_params = {"prominence": prominence, "distance": 3}

    peaks = find_peaks(spint, **find_peaks_params)[0]
    # remove peaks close to baseline level
    peaks = np.setdiff1d(peaks, baseline_index, assume_unique=True)
    peaks = peaks[((spint[peaks] - baseline[peaks]) / noise[peaks]) > min_snr]

    start, end = _find_peak_extension(peaks, baseline_index)
    start, end = _fix_peak_overlap(spint, start, peaks, end)
    start, peaks, end = _normalize_peaks(spint, start, peaks, end)

    # peak centroid and total intensity computation
    # if m[0], ...,  m[n] is the m/z array and i[0], ..., i[n] is the
    # intensity array, for a peak with start and indices k and l respectively,
    # the total intensity A is A = \sum_{j=k}^{l} i[j] and the centroid C,
    # computed as the weighted mean of the m/z is
    # C = \sum_{j=k}^{l} m[j] * i[j] / A
    # If we define the cumulative intensity I_{k} = \sum_{j=0}^{k} i[j]
    # It is easy to see that A = I[l - 1] - I[k - 1]. The same can be done
    # for the centroids defining the weights W[k] = \sum_{j=0}^{k} m[j] * i[j]
    # C = (W[l - 1] - W[k - 1]) / A
    if start.size:
        cumulative_spint = np.cumsum(spint)
        weights = np.cumsum(mz * spint)
        start_cumulative_spint = cumulative_spint[start - 1]
        if start[0] == 0:
            # prevents using the last value from cumulative_spint
            start_cumulative_spint[0] = 0
        total_spint = cumulative_spint[end - 1] - start_cumulative_spint

        start_weight = weights[start - 1]
        if start[0] == 0:
            start_weight[0] = 0
        centroid = (weights[end] - start_weight) / total_spint
    else:
        centroid = np.array([])
        total_spint = np.array([])

    if centroid.size:
        _merge_close_peaks(centroid, total_spint, min_distance)
    return centroid, total_spint


class InvalidPeakException(ValueError):
    """
    Exception raised when invalid indices are used in the construction of
    Peak objects.

    """
    pass


def _find_peak_extension(peaks: np.array, baseline_index: np.array
                         ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Finds the closest baseline points to the left and right of each peak.

    aux function of detect_peaks

    Parameters
    ----------
    peaks : array of indices
    baseline_index : array of indices

    Returns
    -------
    start : array
    end : array

    """
    ext_index = np.searchsorted(baseline_index, peaks)
    ext_index[ext_index >= baseline_index.size] = baseline_index.size - 1
    start = baseline_index[ext_index - 1]
    end = baseline_index[ext_index] + 1
    return start, end


def _fix_peak_overlap(y: np.ndarray, start: np.ndarray, peaks: np.ndarray,
                      end: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Fix boundaries of overlapping peaks.

    aux function of detect_peaks

    Parameters
    ----------
    y : array
    start : array
    peaks : array
    end : array

    Returns
    -------
    start : array
    end : array

    """
    local_min = find_peaks(-y)[0]
    # find overlapping peaks indices
    overlap_mask = end > np.roll(start, -1)
    if overlap_mask.size:
        overlap_mask[-1] = False
    overlap_index = np.where(overlap_mask)[0]
    for k in overlap_index:
        # search local min in the region defined by the overlapping peaks
        ks, ke = np.searchsorted(local_min, [peaks[k], peaks[k + 1]])
        k_min = np.argmin(y[local_min[ks:ke]])
        boundary = local_min[ks + k_min]
        end[k] = boundary + 1
        start[k + 1] = boundary
    return start, end


def _normalize_peaks(y: np.ndarray, start: np.ndarray, peaks: np.ndarray,
                     end: np.ndarray
                     ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Sanity check of peaks.

    Finds the local maximum in the region defined between start and end for each
    peak. If no maximum is found, the peak is removed. Also, corrects the peak
    index using this value that can be slightly shifted if smoothing was
    applied to the signal.

    aux function of detect_peaks.

    Parameters
    ----------
    y : array
    start : array of indices
    peaks : array of indices
    end : array of indices

    Returns
    -------
    start : array of indices
    fixed_peaks : array of indices
    end : array of indices

    """
    local_max = find_peaks(y)[0]
    start_index = np.searchsorted(local_max, start)
    end_index = np.searchsorted(local_max, end, side="right")
    fixed_peaks = np.zeros_like(peaks)
    valid_peaks = np.zeros(peaks.size, dtype=bool)
    for k in range(peaks.size):
        # search local min in the region defined by the overlapping peaks
        local_max_slice = y[local_max[start_index[k]:end_index[k]]]
        if local_max_slice.size > 0:
            k_max = np.argmax(local_max_slice)
            new_peak = local_max[start_index[k] + k_max]
            valid_peaks[k] = (start[k] < new_peak) and (new_peak < end[k])
            fixed_peaks[k] = new_peak
    start = start[valid_peaks]
    end = end[valid_peaks]
    fixed_peaks = fixed_peaks[valid_peaks]
    return start, fixed_peaks, end


def _fill_filter_boundaries(filter_dict: Dict[str, Tuple]):
    """
    Replaces None in the filter boundaries to perform comparisons.

    aux function of get_peak_descriptors
    """
    for k in filter_dict:
        lb, ub = filter_dict[k]
        if lb is None:
            lb = -np.inf
        if ub is None:
            ub = np.inf
        filter_dict[k] = (lb, ub)


def _has_all_valid_descriptors(peak_descriptors: Dict[str, float],
                               filters: Dict[str, Tuple[float, float]]) -> bool:
    """
    Check that the descriptors of a peak are in a valid range.

    aux function of get_peak_descriptors.

    Parameters
    ----------
    peak_descriptors : dict
        mapping of descriptor names to descriptor values.
    filters : dict
        Dictionary from descriptors names to minimum and maximum acceptable
        values.

    Returns
    -------
    is_valid : bool
        True if all descriptors are inside the valid ranges.

    """
    res = True
    for descriptor, (lb, ub) in filters.items():
        d = peak_descriptors[descriptor]
        is_valid = (d >= lb) and (d <= ub)
        if not is_valid:
            res = False
            break
    return res


def _estimate_local_noise(x: np.ndarray, robust: bool = True) -> float:
    r"""
    Estimates noise in a 1D signal. Assumes that the noise is gaussian iid.

    aux function of estimate_noise

    Parameters
    ----------
    x : 1D array
        The size of x must be at least 4. If the size is smaller, the function
        will return 0.
    robust : bool
        If True, estimates the noise using the median absolute deviation. Else
        uses the standard deviation.

    Returns
    -------
    noise : non-negative number.

    """
    d2x = np.diff(x, n=2)
    sorted_index = np.argsort(np.abs(d2x))
    d2x = d2x[sorted_index]
    # if d2x follows a normal distribution ~ N(0, 2*sigma), its sample mean
    # has a normal distribution ~ N(0,  2 * sigma / sqrt(n - 2)) where n is the
    # size of d2x.
    # d2x with high absolute values are removed until this the mean of d2x is
    # lower than its standard deviation.
    # start at 90th percentile and decrease it in each iteration.
    # The loop stops at the 20th percentile even if this condition is not meet
    n_deviations = 3    # dummy values to initialize the loop
    percentile_counter = 9  # start at 90th percentile
    noise_std = 0
    while (n_deviations > 1.0) and (percentile_counter > 2):
        percentile_index = percentile_counter * d2x.size // 10
        # the minimum number of elements required to compute the MAD
        if percentile_index <= 2:
            break
        # dev_threshold = 2 / np.sqrt(percentile - 2)

        if robust:
            noise_std = mad(d2x[:percentile_index], scale="normal")
            noise_mean = np.median(d2x[:percentile_index])
        else:
            noise_std = d2x[:percentile_index].std()
            noise_mean = d2x[:percentile_index].mean()

        # if all the values in d2x are equal, noise_std is equal to zero
        if noise_std > 0:
            n_deviations = abs(noise_mean / noise_std)
        else:
            break
        percentile_counter -= 1
    noise = noise_std / 2
    return noise


def _find_baseline_points(x: np.ndarray, noise: np.ndarray,
                          min_proba: float) -> np.ndarray:
    """
    Finds points flagged as baseline.

    Aux function of estimate_baseline.

    Parameters
    ----------
    x : 1D array
    noise : 1D array
    min_proba : number between 0 and 1

    Returns
    -------
    baseline_index : array of indices.

    """
    extrema = _find_local_extrema(x)
    # check how likely is that the difference observed in each min-max slice
    # can be attributed to noise.
    noise_proba = _estimate_noise_probability(noise, x, extrema)
    # creates a vector with indices where baseline was found
    baseline_index = _build_baseline_index(x, noise_proba, min_proba, extrema)
    return baseline_index


def _find_local_extrema(x: np.ndarray) -> np.ndarray:
    """
    Finds all local minima and maxima in an 1D array.

    aux function of _find_baseline_points.

    Parameters
    ----------
    x : np.ndarray

    Returns
    -------
    np.ndarray : sorted array of local extrema.

    """
    local_max = find_peaks(x)[0]
    local_min = find_peaks(-x)[0]
    if local_max.size:
        # include first and last indices
        extrema = np.hstack([0, local_min, local_max, x.size - 1])
    else:
        extrema = np.array([], dtype=int)
    return np.unique(extrema)


def _estimate_noise_probability(noise: np.ndarray, x: np.ndarray,
                                extrema: np.ndarray) -> np.ndarray:
    """
    Computes the probability that the variation observed in each slice is
    due to noise only.

    Aux function of _find_baseline_points.

    Parameters
    ----------
    noise : 1D array
    x : 1D array
    extrema : array of indices.

    Returns
    -------
    noise_probability : array

    """

    if extrema.size:
        noise_slice = _get_noise_slice_sum_std(noise, extrema[:-1])
        # The difference between maximum and minimum in each slice
        # delta = np.abs(x[np.roll(extrema, -1)] - x[extrema])[:-1]
        x_sum = _get_signal_sum(x, extrema)
        # here we are computing p(abs(sum noise) > delta) assuming a normal
        # distribution
        noise_probability = erfc(x_sum / (noise_slice * np.sqrt(2)))
    else:
        # prevents raising an exception when no extrema had been found
        noise_probability = np.array([])
    return noise_probability


def _get_noise_slice_sum_std(noise: np.ndarray,
                             extrema: np.ndarray) -> np.ndarray:
    """
    Computes the standard deviation of the sum of slices between local maxima.

    aux function of _estimate_noise_probability.

    Parameters
    ----------
    noise : array with the noise level in the signal
    extrema : array with sorted indices of local max/min.

    Returns
    -------
    array

    """
    # the values in noise are an estimation of the standard deviation of the
    # noise. If the noise is iid, the std of the sum is the sum of variances.

    # reshape the extrema indices to compute the sum of elements between
    # consecutive slices, i.e: the sum between of y between extrema[0] and
    # extrema[1], extrema[1] and extrema[2]...
    # The last element is not used
    reduce_ind = (np.vstack([extrema, np.roll(extrema + 1, -1)])
                    .T.reshape(extrema.size * 2)[:-1])
    return np.sqrt(np.add.reduceat(noise ** 2, reduce_ind)[::2])


def _get_signal_sum(x, extrema):
    # we want to compute the sum of x between a, b, where a and b two
    # consecutive local maxima indices. if X is the cumulative sum of x, then
    # the sum between a and b is X[b] - X[a - 1]. The sum must also be relative
    # to the minimum between a and b. Because we are comparing intervals
    # between loca extrema, the intervals are monotonic, and we just need to
    # subtract the min(x[a], x[b]) multiplied by the length of the interval.
    # this code achieves this in a vectorized way
    cum_x = np.cumsum(x)
    ext_shift = np.roll(extrema, -1)
    n_times = ext_shift - extrema + 1
    start_cum_int = cum_x[extrema - 1]
    # first value set to 0 to avoid errors associated with roll
    start_cum_int[0] = 0
    x_sum = cum_x[ext_shift] - start_cum_int
    x_min = np.minimum(x[ext_shift], x[extrema])
    res = x_sum - n_times * x_min
    res = res[:-1]
    return res


def _build_baseline_index(x: np.ndarray, noise_probability: np.ndarray,
                          min_p: float, extrema: np.ndarray):
    """
    builds an array with indices of points flagged as baseline.

    aux function of _find_baseline_points

    Returns
    -------
    array

    """
    # define regions of signal based on noise probability
    is_signal = noise_probability < min_p
    # extend regions of signals to right and left
    is_signal = is_signal | np.roll(is_signal, 1) | np.roll(is_signal, -1)
    baseline_index = list()

    for k in range(extrema.size - 1):
        if not is_signal[k]:
            slice_indices = np.arange(extrema[k], extrema[k + 1] + 1)
            baseline_index.append(slice_indices)

    baseline_index = _include_first_and_last_index(x, baseline_index)
    return baseline_index


def _include_first_and_last_index(x: np.ndarray,
                                  baseline_index: List[np.ndarray]
                                  ) -> np.ndarray:
    """
    adds first and last indices of x to the baseline indices.

    aux function of _build_baseline_index

    Parameters
    ----------
    x : array
    baseline_index : array of indices

    Returns
    -------
    baseline_index : array of indices

    """
    if len(baseline_index):
        stack = list()
        # include first and last indices
        if baseline_index[0][0] != 0:
            stack.append([0])
        stack.extend(baseline_index)
        if baseline_index[-1][-1] != x.size - 1:
            stack.append([x.size - 1])
        baseline_index = np.hstack(stack)
    else:
        baseline_index = np.array([0, x.size - 1], dtype=int)
    return baseline_index


def _merge_close_peaks(mz: np.ndarray, spint: np.ndarray,
                       min_distance: float) -> Tuple[np.ndarray, np.ndarray]:
    dmz = np.diff(mz)
    is_close_mask = (dmz < min_distance) & (np.roll(dmz, -1) > min_distance)
    is_close_mask[-1] = (mz[-1] - mz[-2]) < min_distance    # boundary case
    close_index = np.where(is_close_mask)[0]
    while close_index.size > 0:
        # merge close centroids
        new_spint = spint[close_index] + spint[close_index + 1]
        new_mz = (mz[close_index] * spint[close_index] +
                  mz[close_index + 1] * spint[close_index + 1])
        new_mz /= new_spint
        spint[close_index] = new_spint
        mz[close_index] = new_mz
        # remove merged centroids
        mz = np.delete(mz, close_index + 1)
        spint = np.delete(spint, close_index + 1)

        # repeat until no close peaks are detected
        dmz = np.diff(mz)
        is_close_mask = (dmz < min_distance) & (np.roll(dmz, -1) > min_distance)
        is_close_mask[-1] = (mz[-1] - mz[-2]) < min_distance
        close_index = np.where(is_close_mask)[0]
    return mz, spint



PACKAGE_NAME = "tidyms"
VERSION = "0.3.3"
LICENSE = 'BSD (3-clause)'
AUTHOR = "Bioanalytical Mass Spectrometry Group at CIBION-CONICET"
AUTHOR_EMAIL = "griquelme.chm@gmail.com"
MAINTAINER = "Gabriel Riquelme"
MAINTAINER_EMAIL = AUTHOR_EMAIL
URL = "https://github.com/griquelme/tidyms"
DESCRIPTION = "Tools for working with MS data in metabolomics"

with open("README.md") as fin:
    LONG_DESCRIPTION = fin.read()
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"

CLASSIFIERS = [
    "License :: OSI Approved :: BSD License",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Scientific/Engineering :: Chemistry",
    "Topic :: Scientific/Engineering :: Medical Science Apps."
]

PACKAGES = ["tidyms"]

PYTHON_REQUIRES = ">=3.8"

INSTALL_REQUIRES = [
    "bokeh>=2.4",
    "Cerberus>=1.3",
    "ipython>=8.1",
    "matplotlib>=3.5.1",
    "numpy<=1.21.4",
    "openpyxl>=3.0",
    "pandas>=1.4.1",
    "requests",
    "scikit-learn>=1.0.2",
    "scipy>=1.8",
    "seaborn>=0.11",
    "statsmodels>=0.13",
    "xlrd>=2.0"
]

if __name__ == "__main__":
    from setuptools import setup
    from sys import version_info

    if version_info[:2] < (3, 8):
        msg = "tidyms requires Python >= 3.8."
        raise RuntimeError(msg)

    setup(name=PACKAGE_NAME,
          version=VERSION,
          author=AUTHOR,
          author_email=AUTHOR_EMAIL,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          license=LICENSE,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
          url=URL,
          classifiers=CLASSIFIERS,
          packages=PACKAGES,
          python_requires=PYTHON_REQUIRES,
          install_requires=INSTALL_REQUIRES,
          include_package_data=True)



"""
Utility functions used inside several modules.

"""


import numpy as np
import pandas as pd
from statsmodels.api import OLS, add_constant
from statsmodels.stats.stattools import jarque_bera, durbin_watson
from scipy.stats import spearmanr, median_abs_deviation
import os.path
from typing import Optional, Union
import json

data_type = Union[pd.DataFrame, pd.Series]
reduced_type = Union[pd.Series, float]


def gauss(x: np.ndarray, mu: float, sigma: float,
          amp: float):
    """
    gaussian curve.

    Parameters
    ----------
    x : np.array
    mu : float
    sigma : float
    amp : float

    Returns
    -------
    gaussian : np.array

    """
    gaussian = amp * np.power(np.e, - 0.5 * ((x - mu) / sigma) ** 2)
    return gaussian


def gaussian_mixture(x: np.ndarray, params: np.ndarray
                     ) -> np.ndarray:
    """
    Mixture of gaussian curves.

    Parameters
    ----------
    x : array
    params: np.ndarray
        parameter for each curve the shape of the array is n_curves by
        3. Each row has parameters for one curve (mu, sigma, amp)

    Returns
    -------
    mixture: np.ndarray
        array with gaussian curves. Each row is a gaussian curve. The shape
        of the array is `params.shape[0]` by `x.size`.
    """
    mixture = np.zeros((params.shape[0], x.size))
    for k_row, param in enumerate(params):
        mixture[k_row] = gauss(x, *param)
    return mixture


def normalize(df: pd.DataFrame, method: str,
              feature: Optional[str] = None) -> pd.DataFrame:
    """
    Normalize samples using different methods.

    Parameters
    ----------
    df: pandas.DataFrame
    method: {"sum", "max", "euclidean", "feature"}
        Normalization method. `sum` normalizes using the sum along each row,
        `max` normalizes using the maximum of each row. `euclidean` normalizes
        using the euclidean norm of the row. `feature` normalizes area using
        the value of a specified feature.
    feature: str, optional
        Feature used for normalization in `feature` mode.

    Returns
    -------
    normalized: pandas.DataFrame

    """
    if method == "sum":
        normalized = df.divide(df.sum(axis=1), axis=0)
    elif method == "max":
        normalized = df.divide(df.max(axis=1), axis=0)
    elif method == "euclidean":
        normalized = df.apply(lambda x: x / np.linalg.norm(x), axis=1)
    elif method == "feature":
        normalized = df.divide(df[feature], axis=0)
    else:
        msg = "method must be `sum`, `max`, `euclidean` or `feature`."
        raise ValueError(msg)
    # replace nans generated by division by zero
    normalized[normalized.isna()] = 0
    return normalized


def scale(df: pd.DataFrame, method: str) -> pd.DataFrame:
    """
    scales features using different methods.

    Parameters
    ----------
    df: pandas.DataFrame
    method: {"autoscaling", "rescaling", "pareto"}
        Scaling method. `autoscaling` performs mean centering scaling of
        features to unitary variance. `rescaling` scales data to a 0-1 range.
        `pareto` performs mean centering and scaling using the square root of
        the standard deviation

    Returns
    -------
    scaled: pandas.DataFrame
    """
    if method == "autoscaling":
        scaled = (df - df.mean()) / df.std()
    elif method == "rescaling":
        scaled = (df - df.min()) / (df.max() - df.min())
    elif method == "pareto":
        scaled = (df - df.mean()) / df.std().apply(np.sqrt)
    else:
        msg = "Available methods are `autoscaling`, `rescaling` and `pareto`."
        raise ValueError(msg)
    # replace nans generated when dividing by zero
    scaled[scaled.isna()] = 0
    return scaled


def transform(df: pd.DataFrame, method: str) -> pd.DataFrame:
    """
    perform common data transformations.

    Parameters
    ----------
    df: pandas.DataFrame
    method: {"log", "power"}
        transform method. `log` applies the base 10 logarithm on the data.
        `power`

    Returns
    -------
    transformed: pandas.DataFrame
    """
    if method == "log":
        transformed = df.apply(np.log10)
    elif method == "power":
        transformed = df.apply(np.sqrt)
    else:
        msg = "Available methods are `log` and `power`"
        raise ValueError(msg)
    return transformed


def sample_to_path(samples, path):
    """
    map sample names to raw path if available.

    Parameters
    ----------
    samples : Iterable[str].
        samples names
    path : str.
        path to raw sample data.

    Returns
    -------
    d : dict

    """
    # TODO: this function should accept an extension parameter to prevent
    #   files with the same name but invalid extensions from being used
    available_files = os.listdir(path)
    filenames = [os.path.splitext(x)[0] for x in available_files]
    full_path = [os.path.join(path, x) for x in available_files]
    d = dict()
    for k, name in enumerate(filenames):
        if name in samples:
            d[name] = full_path[k]
    return d

    
def cv(df: data_type, fill_value: Optional[float] = None) -> reduced_type:
    """
    Computes the Coefficient of variation for each column.

    Used by DataContainer objects to compute metrics.

    """
    res = df.std() / df.mean()
    res = _fill_na(res, fill_value)
    return res


def robust_cv(df: data_type, fill_value: Optional[float] = None
              ) -> reduced_type:
    """
    Estimation of the coefficient of variation using the MAD and median.
    Assumes a normal distribution.
    """

    # 1.4826 is used to estimate sigma in an unbiased way assuming a normal
    # distribution for each feature.
    res = mad(df) / df.median()
    res = _fill_na(res, fill_value)
    return res


def mad(df: data_type) -> reduced_type:
    """
    Computes the median absolute deviation for each column. Fill missing
    values with zero.
    """
    # for dataframes with only one row a series of nan is returned. This is
    # to return the same value as std.
    if isinstance(df, pd.DataFrame):
        if df.shape[0] == 1:
            res = pd.Series(data=np.nan, index=df.columns)
        else:
            res = df.apply(median_abs_deviation, scale="normal")
    else:
        res = median_abs_deviation(df, scale="normal")
    return res


def sd_ratio(df1: pd.DataFrame, df2: pd.DataFrame, robust: bool = False,
             fill_value: Optional[float] = None) -> pd.Series:
    """
    Computes the ratio between the standard deviation of the columns of
    DataFrame1 and DataFrame2.

    Used to compute the D-Ratio metric.

    Parameters
    ----------
    df1 : DataFrame with shape (n1, m)
    df2 : DataFrame with shape (n2, m)
    robust : bool
        If True uses the MAD as an estimator of the standard deviation. Else
        computes the sample standard deviation.
    fill_value : Number used to input NaNs.

    Returns
    -------
    ratio : pd.Series

    """
    if robust:
        ratio = mad(df1) / mad(df2)
    else:
        ratio = df1.std() / df2.std()
    ratio = _fill_na(ratio, fill_value)
    return ratio


def detection_rate(df: data_type, threshold: float = 0.0) -> reduced_type:
    """
    Computes the fraction of values in a column above the `threshold`.

    Parameters
    ----------
    df : DataFrame
    threshold : float

    Returns
    -------
    dr : pd.Series

    """
    if isinstance(df, pd.DataFrame):
        n = df.shape[0]
    else:
        n = df.size

    dr = (df > threshold).sum().astype(int) / n
    return dr


def metadata_correlation(y, x, mode: str = "ols"):
    """
    Computes correlation metrics between two variables.

    Parameters
    ----------
    y : array
    x : array
    mode: {"ols", "spearman"}
        `ols` computes r squared, Jarque-Bera test p-value and Durwin-Watson
        statistic from the ordinary least squares linear regression. `spearman`
        computes the spearman rank correlation coefficient.

    Returns
    -------
    dict

    """
    if mode == "ols":
        ols = OLS(y, add_constant(x)).fit()
        r2 = ols.rsquared
        jb = jarque_bera(ols.resid)[1]  # Jarque Bera test p-value
        dw = durbin_watson(ols.resid)   # Durwin Watson statistic
        res = {"r2": r2, "DW": dw, "JB": jb}
    else:
        res = spearmanr(y, x)[0]
    return res


def _fill_na(s: reduced_type, fill_value: Optional[float]):
    if fill_value is None:
        res = s
    elif isinstance(s, pd.Series):
        res = s.fillna(fill_value)
    elif pd.isna(s):
        res = fill_value
    else:
        res = s
    return res


def _find_closest_sorted(x: np.ndarray,
                         xq: Union[np.ndarray, float, int]) -> np.ndarray:
    """
    Find the index in x closest to each xq element. Assumes that x is sorted.

    Parameters
    ----------
    x: numpy.ndarray
        Sorted vector
    xq: numpy.ndarray
        search vector

    Returns
    -------
    ind: numpy.ndarray
        array with the same size as xq with indices closest to x.

    Raises
    ------
    ValueError: when x or xq are empty.
    """

    if isinstance(xq, (float, int)):
        xq = np.array(xq)

    if (x.size == 0) or (xq.size == 0):
        msg = "`x` and `xq` must be non empty arrays"
        raise ValueError(msg)

    ind = np.searchsorted(x, xq)

    if ind.size == 1:
        if ind == 0:
            return ind
        elif ind == x.size:
            return ind - 1
        else:
            return ind - ((xq - x[ind - 1]) < (x[ind] - xq))

    else:
        # cases where the index is between 1 and x.size - 1
        mask = (ind > 0) & (ind < x.size)
        ind[mask] -= (xq[mask] - x[ind[mask] - 1]) < (x[ind[mask]] - xq[mask])
        # when the index is x.size, then the closest index is x.size -1
        ind[ind == x.size] = x.size - 1
        return ind


def find_closest(x: np.ndarray, xq: Union[np.ndarray, float, int],
                 is_sorted: bool = True) -> np.ndarray:
    """
    Search the closest value between two arrays.

    Parameters
    ----------
    x : array
        Array used to search
    xq : array
        query values
    is_sorted : bool, default=True
        If True, assumes that x is sorted.

    Returns
    -------
    array of indices in x
    """
    if is_sorted:
        return _find_closest_sorted(x, xq)
    else:
        sorted_index = np.argsort(x)
        closest_index = _find_closest_sorted(x[sorted_index], xq)
        return sorted_index[closest_index]


def get_filename(full_path: str) -> str:
    """
    get the filename from a full path.

    Parameters
    ----------
    full_path: str

    Returns
    -------
    filename: str`
    """
    return os.path.splitext(os.path.split(full_path)[1])[0]


def is_unique(s: pd.Series):
    s_unique = s.unique()
    return (s.size == s_unique.size) and (s.values == s.unique()).all()


def get_tidyms_path() -> str:
    """
    Returns the path to the directory where datasets and config files are
    stored.

    Returns
    -------
    path : str
    """
    cache_path = os.path.join("~", ".tidyms")
    cache_path = os.path.expanduser(cache_path)
    return cache_path


def create_tidyms_dir():
    tidyms_dir = get_tidyms_path()
    if not os.path.exists(tidyms_dir):
        os.mkdir(tidyms_dir)


def default_settings():
    settings = {
        "bokeh": {
            "apply_theme": True,
            "theme": {
                "attrs": {
                    "Axis": {
                        "axis_label_text_font_style": "bold",
                    },
                }
            },
            "line": {
                "line_width": 1,
                "line_color": "black",
                "line_alpha": 0.8,
            },
            "varea": {
                "fill_alpha": 0.8,
            },
            "palette": {
                "name": "Set3",
                "size": 9,
            },
            "chromatogram": {
                "figure": {
                    "aspect_ratio": 1.5,
                },
                "xaxis": {
                    "axis_label": "Rt [s]",
                },
                "yaxis": {
                    "axis_label": "Intensity [au]",
                }
            },
            "spectrum": {
                "figure": {
                    "aspect_ratio": 1.5,
                },
                "xaxis": {
                    "axis_label": "m/z",
                },
                "yaxis": {
                    "axis_label": "intensity [au]",
                }
            }
        }
    }
    return settings


def get_settings() -> dict:
    """
    Loads the settings into a dictionary object.

    Returns
    -------
    settings : dict

    """
    tidyms_path = get_tidyms_path()
    settings_path = os.path.join(tidyms_path, "settings.json")
    defaults = default_settings()
    exist_user_settings = os.path.isfile(settings_path)
    mode = "r" if exist_user_settings else "w"

    with open(settings_path, mode) as fin:
        if exist_user_settings:
            user_settings = json.load(fin)
            defaults.update(user_settings)
        else:
            json.dump(defaults, fin)
        settings = defaults
    return settings


def is_notebook() -> bool:
    """
    Returns True if the environment is  jupyter notebook.

    Returns
    -------
    bool

    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


create_tidyms_dir()
SETTINGS = get_settings()



"""
Validation functions for Filter, Pipelines and DataContainer
"""


import warnings
from functools import wraps
from inspect import getfullargspec
import cerberus
from typing import Callable
import numpy as np


def is_callable(field, value, error):
    if not hasattr(value, "__call__"):
        msg = "Must be a string or callable"
        error(field, msg)


def is_all_positive(field, value, error):
    if value is not None:
        cond = (value > 0).all()
        if not cond:
            msg = "All of the values must be positive"
            error(field, msg)
# -----------------------------


def validate(params: dict, validator: cerberus.Validator) -> dict:
    """
    Function used to validate parameters.

    Parameters
    ----------
    params: dict
    validator: cerberus.Validator

    Returns
    -------
    dict: Validated and normalized parameters

    Raises
    ------
    ValueError: if any of the parameters are invalid.
    """
    normalized = validator.normalized(params)
    if not validator.validate(normalized):
        msg = ""
        for field, e_msgs in validator.errors.items():
            for e_msg in e_msgs:
                msg += "{}: {}\n".format(field, e_msg)
        raise ValueError(msg)
    return normalized


class ValidatorWithLowerThan(cerberus.Validator):
    def _validate_lower_than(self, other, field, value):
        """
        Tests if a value is lower than the value of other field.

        The rule's arguments are validated against this schema:
        {"type": "string"}
        """
        if (other not in self.document) or (self.document[other] is None):
            return False
        if value >= self.document[other]:
            msg = "{} must be lower than {}".format(field, other)
            self._error(field, msg)

    def _validate_lower_or_equal(self, other, field, value):
        """
        Tests if a value is lower than the value of other field.

        The rule's arguments are validated against this schema:
        {"type": "string"}
        """
        if (other not in self.document) or (self.document[other] is None):
            return False
        if value > self.document[other]:
            msg = "{}, must be lower or equal than {}".format(field, other)
            self._error(field, msg)

    def _validate_is_positive(self, is_positive, field, value):
        """
        Tests if a value is positive

        The rule's arguments are validated against this schema:
        {"type": "boolean"}
        """
        if is_positive and (value is not None) and (value <= 0):
            msg = "Must be a positive number"
            self._error(field, msg)


def validate_data_matrix(df):
    """
    Check type and values for data matrix.

    Check
    Parameters
    ----------
    df : pd.DataFrame.

    Returns
    -------

    """
    float_columns = df.dtypes == float
    # check that all columns have float values
    if not float_columns.all():
        no_float_columns = df.columns[~float_columns]
        msg = "Some columns have non float values: {}."
        msg = msg.format(", ".join(no_float_columns))
        raise TypeError(msg)
    # check that all elements are non-negative
    if (df < 0).sum().sum():
        raise ValueError("Data matrix has negative elements")
    # check nan
    if (df.isna()).sum().sum():
        warnings.warn("Data matrix has NANs. These values should be imputed "
                      "before further analysis")


def validate_feature_metadata(df):
    """
    Check type and values of rt and mz
    Parameters
    ----------
    df

    Returns
    -------
    """
    # rt, mz and class information check
    if "mz" not in df.columns:
        raise KeyError("mz values are required for all features")
    if "rt" not in df.columns:
        raise KeyError("rt values are required for all features")
    # rt, mz and class and data matrix check:
    if (df["mz"] < 0).any():
        raise ValueError("mz values should be greater than zero")
    if (df["rt"] < 0).any():
        raise ValueError("mz values should be greater than zero")


def validate_sample_metadata(df):
    """
    Check sample class.
    Parameters
    ----------
    df

    Returns
    -------
    """
    if "class" not in df.columns:
        raise KeyError("class information are required for all samples")


def validate_data_container(data_matrix, feature_definitions,
                            sample_info):
    validate_data_matrix(data_matrix)
    validate_feature_metadata(feature_definitions)
    validate_sample_metadata(sample_info)
    different_samples = data_matrix.index.difference(sample_info.index)
    if different_samples.size > 0:
        msg = "Samples names should be equal in data matrix and sample info."
        raise ValueError(msg)
    different_features = data_matrix.columns.difference(
        feature_definitions.index)
    if different_features.size > 0:
        msg = "feature names should be equal in data matrix and feature " \
              "definitions."
        raise ValueError(msg)


def validate_blank_corrector_params(params):
    schema = {
        "corrector_classes": {"type": "list", "nullable": True,
                              "schema": {"type": "string"}},
        "process_classes": {"type": "list", "nullable": True,
                            "schema": {"type": "string"}},
        "mode": {"anyof": [{"type": "string",
                            "allowed": ["lod", "loq", "mean", "max"]},
                           {"check_with": is_callable}]},
        "factor": {"type": "number", "is_positive": True},
        "robust": {"type": "boolean"},
        "process_blanks": {"type": "boolean"}
    }
    validator = ValidatorWithLowerThan(schema)
    validate(params, validator)


def validate_prevalence_filter_params(params):
    schema = {
        "lb": {"type": "number", "min": 0, "max": 1, "lower_or_equal": "ub"},
        "ub": {"type": "number", "min": 0, "max": 1},
        "threshold": {"type": "number", "min": 0},
        "intraclass": {"type": "boolean"},
        "process_classes": {"type": "list", "nullable": True,
                            "schema": {"type": "string"}}
    }
    validator = ValidatorWithLowerThan(schema)
    validate(params, validator)


def validate_dratio_filter_params(params):
    schema = {
        "robust": {"type": "boolean"},
        "lb": {"type": "number", "min": 0, "max": 1, "lower_or_equal": "ub"},
        "ub": {"type": "number", "min": 0, "max": 1}
    }
    validator = ValidatorWithLowerThan(schema)
    validate(params, validator)


def validate_dilution_filter_params(params):
    schema = {
        "min_corr": {"type": "number", "min": 0, "max": 1},
        "plim": {"type": "number", "min": 0, "max": 1},
        "mode": {"allowed": ["ols", "spearman"]}
    }
    validator = ValidatorWithLowerThan(schema)
    validate(params, validator)


def validate_variation_filter_params(params):
    schema = {
        "robust": {"type": "boolean"},
        "lb": {"type": "number", "min": 0, "max": 1, "lower_or_equal": "ub"},
        "ub": {"type": "number", "min": 0, "max": 1},
        "intraclass": {"type": "boolean"},
        "process_classes": {"type": "list", "nullable": True,
                            "schema": {"type": "string"}}
    }
    validator = ValidatorWithLowerThan(schema)
    validate(params, validator)


def validate_batch_corrector_params(params):
    schema = {
        "min_qc_dr": {"type": "number", "is_positive": True, "max": 1},
        "frac": {"type": "number", "is_positive": True, "max": 1,
                 "nullable": True},
        "n_qc": {"type": "integer", "nullable": True, "is_positive": True},
        "interpolator": {"type": "string", "allowed": ["splines", "linear"]},
        "process_qc": {"type": "boolean"},
        "threshold": {"type": "number", "min": 0},
        "corrector_classes": {"type": "list", "nullable": True,
                              "schema": {"type": "string"}},
        "process_classes": {"type": "list", "nullable": True,
                            "schema": {"type": "string"}},
        "method": {"allowed": ["additive", "multiplicative"], "type": "string"}
    }
    validator = ValidatorWithLowerThan(schema)
    validate(params, validator)


def validate_descriptors(params):
    schema = {"type": "dict", "keysrules": {"type": "string"},
              "valuesrules": {"check_with": is_callable}}
    if params is not None:
        validator = ValidatorWithLowerThan(schema)
        validate(params, validator)


def validate_filters(params):
    schema = {"type": "dict", "keysrules": {"type": "string"},
              "valuesrules": {"type": "list",
                              "items": [{'type': 'number', "nullable": True},
                                        {'type': 'number', "nullable": True}]
                              }
              }
    if params is not None:
        validator = ValidatorWithLowerThan(schema)
        validate(params, validator)


def validate_detect_peaks_params(params):
    noise_schema = {"min_slice_size": {"is_positive": True, "type": "integer"},
                    "n_slices": {"is_positive": True, "type": "integer"}
                    }
    baseline_schema = {"min_proba": {"is_positive": True, "max": 1.0,
                                     "type": "number"}
                       }
    schema = \
        {"noise_params": {"type": "dict",
                          "schema": noise_schema,
                          "nullable": True
                          },
         "baseline_params": {"type": "dict",
                             "schema": baseline_schema,
                             "nullable": True
                             },
         "smoothing_strength": {"type": "number",
                                "nullable": True,
                                "is_positive": True}
         }
    if params is not None:
        validator = ValidatorWithLowerThan(schema)
        validate(params, validator)


def spectra_iterator_schema(ms_data):
    n_spectra = ms_data.get_n_spectra()
    schema = {
        "ms_level": {
            "type": "integer",
            "min": 1,
        },
        "start": {
            "type": "integer",
            "min": 0,
            "lower_than": "end",
        },
        "end": {
            "type": "integer",
            "max": n_spectra,
            "default": n_spectra
        },
        "start_time": {
            "type": "number",
            "min": 0.0,
            "lower_than": "end_time",
        },
        "end_time": {
            "type": "number",
            "nullable": True,
        }
    }

    defaults = spectra_iterator_defaults(ms_data)
    set_defaults(schema, defaults)
    return schema


def spectra_iterator_defaults(ms_data):
    return dict()


def make_chromatogram_schema(ms_data) -> dict:

    n_spectra = ms_data.get_n_spectra()

    schema = {
        "mz": {
            "empty": False,
            "coerce": np.array,
        },
        "window": {
            "type": "number",
            "is_positive": True,
            "nullable": True
        },
        "accumulator": {
            "type": "string",
            "allowed": ["mean", "sum"]
        },
        "ms_level": {
            "type": "integer",
            "min": 1,
        },
        "start": {
            "type": "integer",
            "min": 0,
            "lower_than": "end",
        },
        "end": {
            "type": "integer",
            "max": n_spectra,
            "nullable": True,
            "default": n_spectra
        },
        "start_time": {
            "type": "number",
            "min": 0.0,
            "lower_than": "end_time",
        },
        "end_time": {
            "type": "number",
            "nullable": True,
        }
    }

    defaults = make_chromatogram_defaults(ms_data)
    set_defaults(schema, defaults)
    return schema


def make_chromatogram_defaults(ms_data):
    instrument = ms_data.instrument

    if instrument == "qtof":
        window = 0.05
    else:   # orbitrap
        window = 0.01

    defaults = {
        "window": {"default": window}
    }
    return defaults


def make_roi_schema(ms_data):
    n_spectra = ms_data.get_n_spectra()
    schema = {
        "tolerance": {
            "type": "number",
            "is_positive": True,
        },
        "max_missing": {
            "type": "integer",
            "min": 0,
        },
        "targeted_mz": {
            "nullable": True,
            "check_with": is_all_positive
        },
        "multiple_match": {
            "allowed": ["closest", "reduce"]
        },
        "mz_reduce": {
            "anyof": [
                {"allowed": ["mean"]},
                {"check_with": is_callable}
            ]
        },
        "sp_reduce": {
            "anyof": [
                {"allowed": ["mean", "sum"]},
                {"check_with": is_callable}
            ]
        },
        "min_intensity": {
            "type": "number",
            "min": 0.0,
            "nullable": True
        },
        "min_length": {
            "type": "integer",
            "min": 1,
            "nullable": True,
        },
        "pad": {
            "type": "integer",
            "min": 0,
            "nullable": True,
        },
        "ms_level": {
            "type": "integer",
            "min": 1,
        },
        "start": {
            "type": "integer",
            "min": 0,
            "lower_than": "end",
        },
        "end": {
            "type": "integer",
            "max": n_spectra,
            "nullable": True,
            "default": n_spectra
        },
        "start_time": {
            "type": "number",
            "min": 0.0,
            "lower_than": "end_time",
        },
        "end_time": {
            "type": "number",
            "nullable": True,
        },
        "min_snr": {
            "type": "number",
            "is_positive": True,
        },
        "min_distance": {
            "type": "number",
            "is_positive": True,
        }
    }
    defaults = make_roi_defaults(ms_data)
    set_defaults(schema, defaults)
    return schema


def make_roi_defaults(ms_data):
    separation = ms_data.separation
    instrument = ms_data.instrument

    if instrument == "qtof":
        tolerance = 0.01
    else:   # orbitrap
        tolerance = 0.005

    if separation == "uplc":
        min_length = 10
    else:   # hplc
        min_length = 20

    defaults = {
        "pad": {
            "default": 2
        },
        "max_missing": {
            "default": 1
        },
        "min_length": {
            "default": min_length
        },
        "tolerance": {
            "default": tolerance
        },
    }

    return defaults


def accumulate_spectra_schema(ms_data):

    n_spectra = ms_data.get_n_spectra()

    schema = {
        "start": {
            "type": "integer",
            "min": 0,
            "lower_than": "end"
        },
        "end": {
            "type": "integer",
            "max": n_spectra,
            "lower_or_equal": "subtract_right"
        },
        "subtract_left": {
            "type": "integer",
            "lower_or_equal": "start",
            "min": 0,
            "nullable": True,
            "default_setter": lambda doc: doc["start"]
        },
        "subtract_right": {
            "type": "integer",
            "max": n_spectra,
            "nullable": True,
            "default_setter": lambda doc: doc["end"]
        },
        "kind": {
            "type": "string"
        },
        "ms_level": {
            "type": "integer",
            "min": 1
        }
    }
    # no default functions is necessary, defaults can be obtained from
    # start and end
    return schema


def set_defaults(schema: dict, defaults: dict):
    for k, v in schema.items():
        if k in defaults:
            if "anyof" in v:
                # TODO: add code for anyof case
                pass
            else:
                v.update(defaults[k])


def validated_ms_data(schema_getter: Callable):
    """
    Validates input for MSData methods and set default values based on
    the `instrument` and `separation` attributes.

    Parameters
    ----------
    schema_getter : callable
        Function that returns the schema used for validation. The function must
        take the MSData object as the only argument.

    Returns
    -------

    """

    def validator_wrapper(func):
        @wraps(func)
        def func_wrapper(*args, **kwargs):
            # build a dictionary with input
            func_argspec = getfullargspec(func)
            func_argnames = func_argspec.args[1:]    # we don't need self
            params = dict(zip(func_argnames, args[1:]))
            params.update(kwargs)
            ms_data = args[0]
            # get schema and defaults
            schema = schema_getter(ms_data)
            # validate input
            validator = ValidatorWithLowerThan(schema)
            validated = validate(params, validator)
            return func(ms_data, **validated)
        return func_wrapper
    return validator_wrapper
